'''Load a specified DNN model and apply it to a dataset and get the sentence-level embeddings
Written by: Shafiq Joty
Date: Mar 10 2016 '''

# keras related
from __future__ import print_function

from keras.models import model_from_json
from keras.models import Graph, Sequential
from keras.layers.core    import Dense, Dropout, Activation
from keras.layers.embeddings import Embedding
from keras.layers.recurrent import LSTM, GRU, SimpleRNN
from keras.preprocessing import sequence
from keras import backend as K

from keras.utils.np_utils import accuracy, to_categorical
from utilities import aidr
from sklearn import metrics

import numpy as np
import optparse
import sys
import re

from six.moves import cPickle




if __name__ == '__main__':


	# parse user input
	parser = optparse.OptionParser("%prog [options]")

	#file related options
	parser.add_option("-t", "--test-file",          dest="test_file",  help="test file to apply the model [default: %default]")
	parser.add_option("-m", "--model-path",         dest="model_path", help="path to the  model [default: %default]")
	parser.add_option("-p", "--vocab-path",         dest="vocab_path", help="path to the  vocab file of the trained model [default: %default]")
	parser.add_option("-d", "--data-dir",           dest="data_dir",   help="directory containing train, test and dev file (on which the model was trained) [default: %default]")
	parser.add_option("-v", "--vocabulary-size",    dest="max_features", type="int", help="vocabulary size of the source model [default: %default]") # emb matrix row size
	parser.add_option("-l", "--max-tweet-length",   dest="maxlen",       type="int", help="maximul tweet length (for fixed size input) [default: %default]") # input size
	parser.add_option("-T", "--model-type",         dest="model_type",   help="uni or bidirectional [default: %default]") # uni, bi-directional


	parser.set_defaults(
    	data_dir        = None
    	,test_file      = "testfile"
    	,model_path     = None
    	,max_features   = 10000
    	,vocab_path     = None
    	,model_type     = 'bidirectional' #bidirectional, unidirectional (default)
    	,maxlen         = 100
	)

	options,args = parser.parse_args(sys.argv)

	# Read out-model vocab
	# Make sure you create the vocab from the same machine you learned your model. 
	# This is required because we didn't use sort to get the vocab.
	# If it was done using sorted vocab, then no need to run this.

	if not options.vocab_path: 
		print ("Creating the vocabs first ")
		options.vocab_path = options.model_path + ".vocab"
		aidr.extract_and_write_vocab(options.data_dir, options.max_features, options.vocab_path)
		print ("Done.")

		
	print ("loading vocab of the trained model from ", options.vocab_path)
	out_vocab_idmap = cPickle.load(open(options.vocab_path, "rb"))

	print('Numberizing test file based on the source domain vocab (i.e., on which the model was trained)...')
	(X_test, y_test), max_features, label_id, ids, sentences, labels = aidr.load_and_numberize_testfile_using_vocab(out_vocab_idmap, options.test_file, get_item_info=1)
	print('............................')

#	ids       = np.array(ids).reshape(len(ids), 1)
#	labels    = np.array(labels).reshape(len(labels), 1)

#	print("Padding sequences....")
	X_test  = sequence.pad_sequences(X_test,  maxlen=options.maxlen)

	try:
		print('Loading model ...')
		model = model_from_json(open(options.model_path + ".json").read())
		model.load_weights(options.model_path)
		print('Done ...')

	except IOError as e:
		print (options.model_path)
		print ("I/O error({0}): {1}".format(e.errno, e.strerror)) 	


	# some sanity checks for unidirectional 
	nb_classes = len(label_id)
	print(len(X_test),      'tweets')
	print(max_features -3,  'vocabulary size')
	print(nb_classes, 'different classes')
	print('............................')


	if options.model_type.lower() != 'bidirectional':
		if model.layers[0].input_dim != max_features:
			print ("Error! Vocab sizes do not match")
			exit(0)

		if model.layers[-2].output_dim != nb_classes:	
			print ("Error! Label sizes do not match")
			exit(0)

		print("not supported for unidirectional yet...")
		exit(0)	
	
	else:# with a Graph model
		get_lstm_layer_output = K.function([model.inputs[i].input for i in model.input_order],
                                   [model.nodes['dropout'].get_output(train=False)])

		output_emb = get_lstm_layer_output([X_test])[0]

		outfile = options.test_file+".emb.csv"
		FW = open(outfile, 'w')
		
		for lid in range(len(ids)):
			e = " ".join(map (str, map (float, output_emb[lid, :] ) ))
			FW.write(ids[lid] + "," + labels[lid] + "," + e + "\n")

		FW.close()

#		y_test = np.array(y_test).reshape(len(y_test), 1)
#		W = np.concatenate((y_test, output_emb), axis=1)
#		np.savetxt(outfile, W, delimiter=',') 

		print("writing output embeddings in csv format (done)")


