'''Load a specified DNN model and apply it to a dataset'''

# keras related
from __future__ import print_function

from keras.models import model_from_json
from keras.models import Graph, Sequential
from keras.layers.core    import Dense, Dropout, Activation
from keras.layers.embeddings import Embedding
from keras.layers.recurrent import LSTM, GRU, SimpleRNN
from keras.preprocessing import sequence
from keras.utils.np_utils import accuracy, to_categorical
from utilities import aidr
from sklearn import metrics

import numpy as np
import optparse
import sys

from six.moves import cPickle




if __name__ == '__main__':


	# parse user input
	parser = optparse.OptionParser("%prog [options]")

	#file related options
	parser.add_option("-t", "--test-file",          dest="test_file",  help="test file to apply the model [default: %default]")
	parser.add_option("-m", "--model-path",         dest="model_path", help="path to the  model [default: %default]")
	parser.add_option("-p", "--vocab-path",         dest="vocab_path", help="path to the  vocab file of the trained model [default: %default]")
	parser.add_option("-d", "--data-dir",           dest="data_dir",   help="directory containing train, test and dev file (on which the model was trained) [default: %default]")
	parser.add_option("-v", "--vocabulary-size",    dest="max_features", type="int", help="vocabulary size of the source model [default: %default]") # emb matrix row size
	parser.add_option("-l", "--max-tweet-length",   dest="maxlen",       type="int", help="maximul tweet length (for fixed size input) [default: %default]") # input size
	parser.add_option("-T", "--model-type",         dest="model_type",   help="uni or bidirectional [default: %default]") # uni, bi-directional


	parser.set_defaults(
    	data_dir        = None
    	,test_file      = "testfile"
    	,model_path     = None
    	,max_features   = 10000
    	,vocab_path     = None
    	,model_type     = 'bidirectional' #bidirectional, unidirectional (default)
    	,maxlen         = 100
	)

	options,args = parser.parse_args(sys.argv)

#	#read out-model vocab
	if not options.vocab_path:
		print ("Creating the vocabs first ")
		options.vocab_path = options.model_path + ".vocab"
		aidr.extract_and_write_vocab(options.data_dir, options.max_features, options.vocab_path)
		print ("Done.")

#	exit(0)
		
	print ("loading vocab of the trained model from ", options.vocab_path)
	out_vocab_idmap = cPickle.load(open(options.vocab_path, "rb"))

	print('Numberizing test file based on the source domain vocab (i.e., on which the model was trained)...')
	(X_test, y_test), max_features, label_id = aidr.load_and_numberize_testfile_using_vocab(out_vocab_idmap, options.test_file)
	print('............................')

#	print (label_id)
#	exit(0)
#	(X_test, y_test), max_features, label_id = aidr.load_and_numberize_testfile(options.data_dir, options.max_features, options.test_file)


#	print("Padding sequences....")
	X_test  = sequence.pad_sequences(X_test,  maxlen=options.maxlen)


	try:
		print('Loading model ...')
		model = model_from_json(open(options.model_path + ".json").read())
		model.load_weights(options.model_path)
		print('Done ...')

	except IOError as e:
		print (options.model_path)
		print ("I/O error({0}): {1}".format(e.errno, e.strerror)) 	


	# some sanity checks for unidirectional 
	nb_classes = len(label_id)
	print(len(X_test),      'test  tweets')
	print(max_features -3,  'vocabulary size')
	print(nb_classes, 'different classes')
	print('............................')

	if options.model_type.lower() != 'bidirectional':
		if model.layers[0].input_dim != max_features:
			print ("Error! Vocab sizes do not match")
			exit(0)

		if model.layers[-2].output_dim != nb_classes:	
			print ("Error! Label sizes do not match")
			exit(0)

		y_pred = model.predict_classes(X_test)
	
	else:
		y_prob = model.predict({'input': X_test})['output']
		y_pred = np.argmax(y_prob, axis=1)
	
	y_test = np.array(y_test)
	acc2 = metrics.accuracy_score(y_test, y_pred)
	print("Raw Accuracy:", acc2)

	#get label ids in sorted
	class_labels = sorted(label_id, key=label_id.get)
	class_ids    = sorted(label_id.values())

	np.set_printoptions(threshold=np.nan)
	print (y_pred)

	print (metrics.classification_report(y_test, y_pred, labels=class_ids, target_names=class_labels) )
	print ("Confusion Matrix:\n", metrics.confusion_matrix(y_test, y_pred) )

	mic_p, mic_r, mic_f, sup = metrics.precision_recall_fscore_support(y_test, y_pred, average='micro')
	mac_p, mac_r, mac_f, sup = metrics.precision_recall_fscore_support(y_test, y_pred, average='macro')

	print (" micro pre: " + str (mic_p) + " rec: " + str (mic_r) + " f-score: " + str (mic_f))
	print (" macro pre: " + str (mac_p) + " rec: " + str (mac_r) + " f-score: " + str (mac_f))
	print (" final: %f (%f)", mac_f, mic_f)








