### <- set paths - >
DATA_PATH="../../data/input_to_DNNs/QL/"
declare -a DATA_DIR=("$DATA_PATH/0/" "$DATA_PATH/1/" "$DATA_PATH/2/" "$DATA_PATH/3/" "$DATA_PATH/4/")

#"$DATA_PATH/MRDA/" 

RNN_SCR="dnn_scripts/rnns_aidr.py"
EXP_DIR="experiments/rnn_results/QL/"
MODEL_DIR="saved_models/rnn_models/QL/"

mkdir -p $MODEL_DIR
mkdir -p $EXP_DIR

###<- Set general DNN settings ->
dr_ratios=(0.0 0.2)   #dropout_ratio
mb_sizes=(16 32)   #minibatch-size
vocab_sizes=(80 85 90) #how many words in percentage for vocabulary

init_type="random" #instead of word embedding?

### <- set RNN settings ->
declare -a MODEL_TYPES=("bidirectional")
hidden_sizes=(150 200) #no of units in the hidden layer

for data in "${DATA_DIR[@]}"; do
	for model_type in "${MODEL_TYPES[@]}"; do

		d_set=$(echo $data | rev | cut -d '/' -f 2 | rev);
		log="$EXP_DIR/RNN-$model_type.$d_set.log";
		echo "INFORMATION: RNN-$model_type on $d_set" >> "$log";

		for ratio in ${dr_ratios[@]}; do
			for mb in ${mb_sizes[@]}; do

				for h_size in ${hidden_sizes[@]}; do
					for vocab in ${vocab_sizes[@]}; do

						echo "INFORMATION: dropout_ratio=$ratio minibatch-size=$mb hidden-size=$h_size vocab=$vocab" >> $log;
						echo "----------------------------------------------------------------------" >> $log;

						python $RNN_SCR --model-type $model_type --data-dir=$data --model-dir=$MODEL_DIR --init-type=$init_type\
						--vocabulary-size=$vocab --dropout_ratio=$ratio -C 1 --minibatch-size=$mb --vocabulary-size=$vocab\
						-s $h_size -D $d_set >>$log
						wait

						echo "----------------------------------------------------------------------" >> $log;

					done
				done
			done
		done 
	done	
done
